import inspect
import os
import sys
from datetime import datetime
from json import JSONEncoder
from typing import Any

from .customTypes import COLORS


def getCallerFilePath(stack = None) -> str:
    """
    Returns the absolute filepath of the caller of the parent function
    """
    if stack is None:
        stack = inspect.stack()
    # return os.path.abspath(stack[2][1])
    if len(stack) < 3:
        return os.path.abspath(stack[-1].filename)
    return os.path.abspath(stack[2].filename)

def getCallerFunctionName(stack = None) -> str:
    """
    Returns the name of the function that called this one, including the class name if the function is a method
    """
    if stack is None:
        stack = inspect.stack()
    if len(stack) < 3:
        return "<module>"
    caller = stack[2]
    caller_name = caller.function
    if caller_name == "<module>":
        return "<module>"

    parents = getAllParents(caller.filename, caller.lineno)[::-1]
    if len(parents) <= 0:
        return caller_name
    if caller_name == parents[-1]:
        return '.'.join(parents)
    else:
        return '.'.join(parents) + '.' + caller_name


def getCallerInfo():
    stack = inspect.stack()
    return getCallerFilePath(stack), getCallerFunctionName(stack)

def getTime():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


def replaceNewLine(string : str, indent : int = 33):
    return string.replace('\n', '\n' + (' ' * indent) + '| ')


def centerString(string : str, length : int):
    return string.center(length)

def string2type(string : str):
    match string.lower().strip():
        case 'int':
            return int
        case 'str':
            return str
        case 'float':
            return float
        case 'bool':
            return bool
        case 'list':
            return list
        case 'dict':
            return dict
        case 'tuple':
            return tuple
        case 'set':
            return set
        case 'none':
            return None
        case _:
            return Any

def type2string(t : Any):
    match t:
        case None:
            return 'None'
        case _:
            return t.__name__

def getFunctionArguments(func) -> dict[str, list[type]]:
    signature =  str(inspect.signature(func))
    signature = signature[1:-1] # Remove parenthesis
    signature = signature.split(',')

    result = {}
    for arg in signature:
        arg = arg.strip()
        arg = arg.split('=')[0]
        arg = arg.split(':')
        arg = [elem.strip() for elem in arg]
        if len(arg) == 1:
            result[arg[0]] = [Any]
            continue
        types = arg[1]
        result[arg[0]] = [string2type(t) for t in types.split('|')]

    return result

def strictTypeCheck(func):
    def wrapper(*args, **kwargs):
        # Get the function arguments
        func_args = getFunctionArguments(func)

        # Check the arguments
        for i, arg in enumerate(args):
            types = func_args[list(func_args.keys())[i]]
            if type(arg) not in types and Any not in types:
                raise TypeError(f"Argument {i} of function {func.__name__} must be of type ("+", ".join(map(type2string, types)) + ")")

        for key, value in kwargs.items():
            types = func_args[key]
            if type(value) not in types and Any not in types:
                raise TypeError(f"Argument {key} of function {func.__name__} must be of type ("+", ".join(map(type2string, types)) + ")")

        return func(*args, **kwargs)

    return wrapper

def countLinesLength(string : str) -> list[int]:
    lines = string.split('\n')
    return [len(line) for line in lines]

def splitLongString(string : str, length : int = 100) -> str:
    """Split a long string into multiple lines, on spaces"""
    result = []
    if len(string) <= length:
        return string

    lines = [line.split(' ') for line in string.split('\n')]

    for line in lines:
        current_line = []
        for word in line:
            if len(word) > length:
                raise ValueError("A word is longer than the maximum length")
            if len(' '.join(current_line)) + len(word) > length:
                result.append(' '.join(current_line))
                current_line = [word]
            else:
                current_line.append(word)
        result.append(' '.join(current_line))
    return '\n'.join(result)

class CustomJSONEncoder(JSONEncoder):
    def default(self, o):
        # if we serialize an enum, just return the name
        if hasattr(o, '_name_'):
            return o._name_

        if hasattr(o, '__dict__'):
            return o.__dict__
        elif hasattr(o, '__str__'):
            return str(o)
        else:
            return super().default(o)




def getAllParents(filepath, lineno):
    """
    Get all parent classes of a class or method, based on indentation in the file
    """

    # Read
    with open(filepath, 'r', encoding="utf-8") as f:
        lines = f.readlines()

    # Get the line
    line = lines[lineno-1]

    # Get the indentation
    indentation = len(line) - len(line.lstrip())

    # Get the parent classes
    parents = []
    for i in range(lineno-1, 0, -1):
        line = lines[i]
        if len(line) - len(line.lstrip()) < indentation:
            indentation = len(line) - len(line.lstrip())
            if "class" in line:
                parents.append(line.strip()[:-1].split(' ')[1]) # Remove the ':'
            elif "def" in line:
                parents.append(line.strip()[:-1].split(' ')[1].split('(')[0])

    return parents

def colorize(color : COLORS, string : str):
    return f"{color}{string}{COLORS.RESET}"


def getExecutableFormatted() -> str:
    executable = sys.executable.split(os.sep)[-1]
    return f"{executable} {sys.argv[0]}" if 'python' in executable else executable



import threading
from datetime import datetime
from json import dumps
from typing import Any, Callable

from .customTypes import (COLORS, LEVELS, SENSITIVE_LEVELS, TERMINAL_TARGETS,
                          LoggerConfig, Module, Target)
from .utils import (CustomJSONEncoder, centerString, colorize, getCallerInfo,
                    getExecutableFormatted, getTime, replaceNewLine,
                    splitLongString, strictTypeCheck)


class UnexpectedError(Exception): ...

class Logger:

    __instance = None # type: Logger|None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super(Logger, cls).__new__(cls)
            cls.__instance.init()

        return cls.__instance

    def init(self):
        self.config = LoggerConfig()

        #configuring default target
        defaultTargget = Target(TERMINAL_TARGETS.STDOUT)
        defaultTargget["level"] = LEVELS.INFO
        defaultTargget["sensitiveMode"] = SENSITIVE_LEVELS.HIDE

#---------------------------------------- Internal methods ----------------------------------------

    @strictTypeCheck
    def __print(self, level : LEVELS, msg : Any, callerInfo : tuple[str, str]):
        for target in Target.list():
            self.__printInTarget(level, msg, callerInfo, target)

    @strictTypeCheck
    def __printInTarget(self, level : LEVELS, msg : Any, callerInfo : tuple[str, str], target : Target):
        if not target["level"] <= level:
            return
        result = ""

        # add the current time
        result += self.__LogElementTime(target)

        # add the process name if needed
        result += self.__LogElementProcessName(target)

        # add the thread name if needed
        result += self.__LogElementThreadName(target)

        # add the level of the message
        result += self.__LogElementLevel(level, target)

        # add the module name if needed
        result += self.__LogElementModule(callerInfo, target)

        # add the message
        result += self.__LogElementMessage(msg, callerInfo)

        result = self.__parseSensitive(result, target)
        target(result+"\n")

    def __LogElementTime(self, target):
        if target.type == Target.Type.TERMINAL:
            return f"[{COLORS.BLUE}{getTime()}{COLORS.RESET}]"
        else:
            # if the target is a file, we don't need to color the output
            return f"[{getTime()}]"

    def __LogElementProcessName(self, target):
        if self.config.showProcessName:
            if target.type == Target.Type.TERMINAL:
                return f" [{COLORS.CYAN}{centerString(getExecutableFormatted(), 20)}{COLORS.RESET}]"
            else:
                return f" [{centerString(getExecutableFormatted(), 20)}]"
        return ""

    def __LogElementThreadName(self, target):
        if self.config.showThreadsName:
            name = centerString(threading.current_thread().name, 30)
            if target.type == Target.Type.TERMINAL:
                return f" [ {COLORS.CYAN}{name}{COLORS.RESET} ]"
            else:
                return f" [ {name} ]"
        return ""

    def __LogElementLevel(self, level, target):
        if target.type == Target.Type.TERMINAL:
            return f" [{level.color()}{level}{COLORS.RESET}]"
        else:
            return f" [{level}]"

    def __LogElementModule(self, callerInfo, target):
        result = ""
        if Module.exist(*callerInfo):
            for module in Module.get(*callerInfo).getCompletePath():
                if target.type == Target.Type.TERMINAL:
                    result += f" [ {colorize(COLORS.BLUE, centerString(module, 15))} ]"
                else:
                    result += f" [ {centerString(module, 15)} ]"
        return result

    def __LogElementMessage(self, msg, callerInfo):
        if isinstance(msg, (int, float, bool)):
            msg = str(msg)
        elif isinstance(msg, str):
            msg = splitLongString(msg, 150)
        else:
            msg = dumps(msg, indent=4, cls=CustomJSONEncoder)

        return f" {replaceNewLine(msg, 33 + (20 if Module.exist(*callerInfo) else 0))}"

    @strictTypeCheck
    def __printMessageInTarget(self, msg : str, color : COLORS, target : Target):
        msg = self.__parseSensitive(msg, target)
        if target.type == Target.Type.TERMINAL:
            target(f"{color}{msg}{COLORS.RESET}")
        else:
            target(msg+"\n")

    @strictTypeCheck
    def __printMessage(self, msg : str, color : COLORS):
        for target in Target.list():
            self.__printMessageInTarget(msg, color, target)

    @strictTypeCheck
    def __parseSensitive(self, msg : str, target : Target) -> str:
        match target["sensitiveMode"]:
            case SENSITIVE_LEVELS.HIDE:
                for sensitive in self.config.sensitiveDatas:
                    msg = msg.replace(sensitive, "*" * len(sensitive))
                return msg
            case SENSITIVE_LEVELS.SHOW:
                return msg
            case _:
                raise ValueError(f"Unknown sensitive mode: {target['sensitiveMode']}")

#---------------------------------------- Logging methods -----------------------------------------

    @staticmethod
    def deepDebug(msg : Any, callerInfo = None):
        if callerInfo is None:
            callerInfo = getCallerInfo()
        Logger.getInstance().__print(LEVELS.DEEP_DEBUG, msg, callerInfo)

    @staticmethod
    def debug(msg : Any, callerInfo = None):
        if callerInfo is None:
            callerInfo = getCallerInfo()
        Logger.getInstance().__print(LEVELS.DEBUG, msg, callerInfo)

    @staticmethod
    def info(msg : Any, callerInfo = None):
        if callerInfo is None:
            callerInfo = getCallerInfo()
        Logger.getInstance().__print(LEVELS.INFO, msg, callerInfo)

    @staticmethod
    def warning(msg : Any, callerInfo = None):
        if callerInfo is None:
            callerInfo = getCallerInfo()
        Logger.getInstance().__print(LEVELS.WARNING, msg, callerInfo)

    @staticmethod
    def error(msg : Any, callerInfo = None):
        if callerInfo is None:
            callerInfo = getCallerInfo()
        Logger.getInstance().__print(LEVELS.ERROR, msg, callerInfo)

    @staticmethod
    def critical(msg : Any, callerInfo = None):
        if callerInfo is None:
            callerInfo = getCallerInfo()
        Logger.getInstance().__print(LEVELS.CRITICAL, msg, callerInfo)

    @staticmethod
    @strictTypeCheck
    def message(msg : Any, color : COLORS = COLORS.NONE):
        Logger.getInstance().__printMessage(msg, color)

#---------------------------------------- Configuration methods -----------------------------------

    @staticmethod
    def getInstance():
        return Logger.__instance

    @staticmethod
    @strictTypeCheck
    def setLevel(targetName: str, level : LEVELS):
        target = Target.get(targetName)
        target["level"] = level

    @staticmethod
    @strictTypeCheck
    def setSensitiveMode(targetName: str, mode : SENSITIVE_LEVELS):
        target = Target.get(targetName)
        target["sensitiveMode"] = mode

        if mode == SENSITIVE_LEVELS.SHOW:
            Logger.getInstance().__printMessageInTarget("Sensitive mode was disable, this file may contain sensitive information, please do not share it with anyone", COLORS.YELLOW, target)

    @staticmethod
    def setModule(name : str):
        callerInfo = getCallerInfo()
        if not name:
            Module.delete(*callerInfo)
        elif len(name) > 15:
            raise ValueError("Module name should be less than 15 characters")
        else:
            Module.new(name, *callerInfo)


    @staticmethod
    @strictTypeCheck
    def showThreadsName(value : bool = True):
        Logger.getInstance().config.showThreadsName = value

    @staticmethod
    @strictTypeCheck
    def showProcessName(value : bool = True):
        Logger.getInstance().config.showProcessName = value


    @staticmethod
    @strictTypeCheck
    def addTarget(targetFunc : Callable[[str], None] | str | Target | TERMINAL_TARGETS, level : LEVELS = LEVELS.INFO, sensitiveMode : SENSITIVE_LEVELS = SENSITIVE_LEVELS.HIDE) -> str:
        target = None #type: Target|None
        if isinstance(targetFunc, str):
            target = Target.fromFile(targetFunc)
        elif isinstance(targetFunc, Target):
            target = targetFunc
        else:
            target = Target(targetFunc)
        Logger.setLevel(target.name, level)
        Logger.setSensitiveMode(target.name, sensitiveMode)
        return target.name

    @staticmethod
    @strictTypeCheck
    def removeTarget(targetName : str):
        Target.unregister(targetName)

    @staticmethod
    @strictTypeCheck
    def addSensitiveData(data : Any):
        Logger.getInstance().config.sensitiveDatas.append(data)


    @staticmethod
    @strictTypeCheck
    def reset():
        Target.clear()
        Logger.getInstance().config.clear()

        if not Logger.__instance:
            raise UnexpectedError("Logger instance does not exist")

        #configuring default target
        defaultTargget = Target(TERMINAL_TARGETS.STDOUT)
        defaultTargget["level"] = LEVELS.INFO
        defaultTargget["sensitiveMode"] = SENSITIVE_LEVELS.HIDE

def deepDebug(msg : Any):
    Logger.deepDebug(msg, getCallerInfo())

def debug(msg : Any):
    Logger.debug(msg, getCallerInfo())

def info(msg : Any):
    Logger.info(msg, getCallerInfo())

def warning(msg : Any):
    Logger.warning(msg, getCallerInfo())

def error(msg : Any):
    Logger.error(msg, getCallerInfo())

def critical(msg : Any):
    Logger.critical(msg, getCallerInfo())

@strictTypeCheck
def message(msg : Any, color : COLORS = COLORS.NONE):
    """
    Print a message to the standard output, in yellow color\n
    This method should be used before any other method\n
    It is used to pass information to the user about the global execution of the program
    """
    Logger.message(msg, color)

@strictTypeCheck
def deepDebugFunc(useChrono : bool = False):
    """
    Decorator to print deep debug messages before and after the function call
    usage:
    ```python
    @deep_debug_func(useChrono=False)
    def my_function(arg1, arg2, kwarg1=None):
        return arg1+arg2

    my_function("value1", "value2", kwarg1="value3")
    ```
    will print:
    ```log
    [datetime] [   DEBUG   ] Calling my_function with\n\t\t\t   | args: (value1, value2)\n\t\t\t   | kwargs: {'kwarg1': 'value3'}
    [datetime] [   DEBUG   ] Function my_function returned "value1value2"
    ```

    note: this decorator does nothing if the Logger level is not set to deep debug
    """
    @strictTypeCheck
    def pre_wrapper(func : Callable):
        @strictTypeCheck
        def wrapper(*args, **kwargs):
            deepDebug(f"Calling {func.__name__} with\nargs: {args}\nkwargs: {kwargs}")
            if useChrono:
                start = datetime.now()
            result = func(*args, **kwargs)
            if useChrono:
                end = datetime.now()
                tDelta = str(end-start).split(".")[0]
                deepDebug(f"Function {func.__name__} took {tDelta} to execute and returned \"{result}\"")
            else:
                deepDebug(f"Function {func.__name__} returned \"{result}\"")
            return result
        return wrapper
    return pre_wrapper

@strictTypeCheck
def debugFunc(useChrono : bool = False):
    """
    Decorator to print deep debug messages before and after the function call
    usage:
    ```python
    @deep_debug_func
    def my_function(arg1, arg2, kwarg1=None):
        return arg1+arg2

    my_function("value1", "value2", kwarg1="value3")
    ```
    will print:
    ```log
    [datetime] [   DEBUG   ] Calling my_function with\n\t\t\t   | args: (value1, value2)\n\t\t\t   | kwargs: {'kwarg1': 'value3'}
    [datetime] [   DEBUG   ] Function my_function returned "value1value2"
    ```

    note: this decorator does nothing if the Logger level is not set to debug or deep debug
    """
    @strictTypeCheck
    def pre_wrapper(func : Callable):
        @strictTypeCheck
        def wrapper(*args, **kwargs):
            debug(f"Calling {func.__name__} with\nargs: {args}\nkwargs: {kwargs}")
            if useChrono:
                start = datetime.now()
            try:
                result = func(*args, **kwargs)
            except Exception as e:
                error(f"An error occurred in function {func.__name__}: {e.__class__.__name__} - {e}")
                raise e
            if useChrono:
                end = datetime.now()
                debug(f"Function {func.__name__} took {end-start} to execute and returned \"{result}\"")
            else:
                debug(f"Function {func.__name__} returned \"{result}\"")
            return result
        return wrapper
    return pre_wrapper

@strictTypeCheck
def chrono(func : Callable):
    """
    Decorator to print the execution time of a function
    usage:
    ```python
    @chrono
    def my_function(arg1, arg2, kwarg1=None):
        return arg1+arg2

    my_function("value1", "value2", kwarg1="value3")
    ```
    will print:
    ```log
    [datetime] [   DEBUG   ] Function my_function took 0.0001s to execute
    ```
    """
    @strictTypeCheck
    def wrapper(*args, **kwargs):
        start = datetime.now()
        result = func(*args, **kwargs)
        end = datetime.now()
        debug(f"Function {func.__name__} took {end-start} to execute")
        return result
    return wrapper

# create the instance of the logger
Logger()



import sys
import threading
from enum import Enum
from typing import Any, Callable


class Module:
    __instances = {}  # type: dict[tuple[str|None, str|None], Module]
    def __init__(self, name : str, parent : 'Module|None' = None, file : str|None = None, function : str|None = None):
        self.parent = parent
        self.name = name
        self.file = file
        self.function = function

        Module.__instances[(self.file, self.function)] = self

    def getCompleteName(self) -> str:
        if self.parent is None:
            return self.name
        return f'{self.parent.getCompleteName()}.{self.name}'

    def getCompletePath(self) -> list[str]:
        if self.parent is None:
            return [self.name]
        return self.parent.getCompletePath() + [self.name]

    @staticmethod
    def get(filename : str, function : str) -> 'Module':
        if Module.exist(filename, function):
            return Module.__instances[(filename, function)]
        else:
            raise ValueError(f"No module found for file {filename} and function {function}")

    @staticmethod
    def exist(filename : str, function : str) -> bool:
        return (filename, function) in Module.__instances

    @staticmethod
    def delete(filename : str, function : str):
        if Module.exist(filename, function):
            del Module.__instances[(filename, function)]
        else:
            raise ValueError(f"No module found for file {filename} and function {function}")

    @staticmethod
    def getByName(name : str) -> 'Module':
        for module in Module.__instances.values():
            if module.getCompleteName() == name:
                return module
        raise ValueError(f"No module found for name {name}")

    @staticmethod
    def existByName(name : str) -> bool:
        return any(
            module.getCompleteName() == name
            for module in Module.__instances.values()
        )

    @staticmethod
    def deleteByName(name : str):
        if not Module.existByName(name):
            raise ValueError(f"No module found for name {name}")
        module = Module.getByName(name)
        del Module.__instances[(module.file, module.function)]


    @staticmethod
    def clear():
        Module.__instances = {}

    @staticmethod
    def new(name : str, file : str|None = None, function : str|None = None) -> 'Module':
        if Module.existByName(name):
            existing = Module.getByName(name)
            if file == existing.file and function == existing.function:
                return existing
            else:
                raise ValueError(f"Module {name} already exists with file {existing.file} and function {existing.function}")

        if '.' in name:
            parentName, moduleName = name.rsplit('.', 1)
            if not Module.existByName(parentName):
                raise ValueError(f"No module found for name {parentName}")
            #get the parent module
            parent = Module.getByName(parentName)
            return Module(moduleName, parent, file, function)
        return Module(name, None, file, function)


class COLORS(Enum):
    """
    usage:
    ```python
    print(COLORS.RED + "This is red text" + COLORS.RESET)
    print(COLORS.GREEN + "This is green text" + COLORS.RESET)
    print(COLORS.YELLOW + "This is yellow text" + COLORS.RESET)
    ```
    """
    RED = '\033[91m'
    DARK_RED = '\033[91m\033[1m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    MAGENTA = '\033[96m'
    CYAN = '\033[96m'
    RESET = '\033[0m'
    NONE = ''

    def __str__(self):
        return self.value

    def __add__(self, other):
        """
        Allow to concatenate a string with a color, example:
        ```python
        print(COLORS.RED + "This is red text" + COLORS.RESET)
        ```
        or using an f-string:
        ```python
        print(f"{COLORS.RED}This is red text{COLORS.RESET}")
        ```
        """
        return f"{self}{other}"

    def __radd__(self, other):
        """
        Allow to concatenate a string with a color, example:
        ```python
        print(COLORS.RED + "This is red text" + COLORS.RESET)
        ```
        or using an f-string:
        ```python
        print(f"{COLORS.RED}This is red text{COLORS.RESET}")
        ```
        """
        return f"{other}{self}"

    def __repr__(self):
        return self.value

class LEVELS(Enum):
    """
    ## list of levels:
    - DEEP_DEBUG:   this level is used to print very detailed information, it may contain sensitive information
    - DEBUG:        this level is used to print debug information, it may contain sensitive information
    - INFO:         this level is used to print information about the normal execution of the program
    - WARNING:      this level is used to print warnings about the execution of the program (non-blocking, but may lead to errors)
    - ERROR:        this level is used to print errors that may lead to the termination of the program
    - CRITICAL:     this level is used to print critical errors that lead to the termination of the program, typically used in largest except block
    """

    DEEP_DEBUG = 0  # this level is used to print very detailed information, it may contain sensitive information
    DEBUG = 1       # this level is used to print debug information, it may contain sensitive information
    INFO = 2        # this level is used to print information about the normal execution of the program
    WARNING = 3     # this level is used to print warnings about the execution of the program (non-blocking, but may lead to errors)
    ERROR = 4       # this level is used to print errors that may lead to the termination of the program
    CRITICAL = 5    # this level is used to print critical errors that lead to the termination of the program, typically used in largest except block


    @staticmethod
    def from_string(level : str) -> 'LEVELS':
        match level.lower():
            case 'deep_debug':
                return LEVELS.DEEP_DEBUG
            case 'debug':
                return LEVELS.DEBUG
            case 'info':
                return LEVELS.INFO
            case 'warning':
                return LEVELS.WARNING
            case 'error':
                return LEVELS.ERROR
            case 'critical':
                return LEVELS.CRITICAL
            case _:
                return LEVELS.INFO

    def __str__(self) -> str:
        """
        Return the string representation of the level, serialized to 10 characters (centered with spaces)
        """
        match self:
            case LEVELS.DEEP_DEBUG:
                return '  DEBUG   '
            case LEVELS.DEBUG:
                return '  DEBUG   '
            case LEVELS.INFO:
                return '   INFO   '
            case LEVELS.WARNING:
                return ' WARNING  '
            case LEVELS.ERROR:
                return '  ERROR   '
            case LEVELS.CRITICAL:
                return ' CRITICAL '

    def __int__(self):
        return self.value

    def __le__(self, other : 'LEVELS'):
        return self.value <= other.value

    def color(self) -> COLORS:
        match self:
            case LEVELS.DEEP_DEBUG:
                return COLORS.BLUE
            case LEVELS.DEBUG:
                return COLORS.BLUE
            case LEVELS.INFO:
                return COLORS.GREEN
            case LEVELS.WARNING:
                return COLORS.YELLOW
            case LEVELS.ERROR:
                return COLORS.RED
            case LEVELS.CRITICAL:
                return COLORS.DARK_RED

class SENSITIVE_LEVELS(Enum):
    HIDE = 10
    SHOW = 11

    @staticmethod
    def from_string(level : str) -> 'SENSITIVE_LEVELS':
        match level.lower():
            case 'hide':
                return SENSITIVE_LEVELS.HIDE
            case 'show':
                return SENSITIVE_LEVELS.SHOW
            case _:
                return SENSITIVE_LEVELS.HIDE

    @staticmethod
    def from_bool(value : bool) -> 'SENSITIVE_LEVELS':
        return SENSITIVE_LEVELS.SHOW if value else SENSITIVE_LEVELS.HIDE

class TERMINAL_TARGETS(Enum):
    STDOUT = 30
    STDERR = 31

    def __str__(self) -> str:
        match self:
            case TERMINAL_TARGETS.STDOUT:
                return 'stdout'
            case TERMINAL_TARGETS.STDERR:
                return 'stderr'

    @staticmethod
    def from_string(target : str) -> 'TERMINAL_TARGETS':
        match target.lower():
            case 'stdout':
                return TERMINAL_TARGETS.STDOUT
            case 'stderr':
                return TERMINAL_TARGETS.STDERR
            case _:
                raise ValueError(f"Invalid terminal target: {target}")

class Target:
    __instances = {} #type: dict[str, Target]

    class Type(Enum):
        FILE = 20
        TERMINAL = 21

        def __str__(self) -> str:
            match self:
                case Target.Type.FILE:
                    return 'file'
                case Target.Type.TERMINAL:
                    return 'terminal'

    def __new__(cls, target : Callable[[str], None] | TERMINAL_TARGETS, name : str|None = None):
        if name is None:
            if isinstance(target, TERMINAL_TARGETS):
                name = name if name is not None else str(target)
            elif hasattr(target, '__name__'):
                name = target.__name__
            else:
                raise ValueError("The target must be a function or a TERMINAL_TARGETS; use Target.fromFile(file) to create a file target")
        if target in cls.__instances:
            return cls.__instances[name]
        instance = super().__new__(cls)
        cls.__instances[name] = instance
        return instance

    def __init__(self, target : Callable[[str], None] | TERMINAL_TARGETS, name : str|None = None):

        if isinstance(target, str):
            raise ValueError("The target must be a function or a TERMINAL_TARGETS; use Target.fromFile(file) to create a file target")

        if isinstance(target, TERMINAL_TARGETS):
            match target:
                case TERMINAL_TARGETS.STDOUT:
                    self.target = sys.stdout.write
                case TERMINAL_TARGETS.STDERR:
                    self.target = sys.stderr.write
            self.__type = Target.Type.TERMINAL
            self.__name = name if name is not None else str(target)
        else:
            self.__type = Target.Type.FILE
            self.__name = name if name is not None else target.__name__
            self.target = target


        self.properties = {} #type: dict[str, Any]
        self.__lock = threading.Lock()

    @staticmethod
    def fromFile(file : str) -> 'Target':
        def writeToFile(string : str):
            with open(file, 'a', encoding="utf-8") as f:
                f.write(string)
        with open(file, 'w', encoding="utf-8") as f: # clear the file
            f.write('')
        return Target(writeToFile, file)

    def __call__(self, string : str):
        with self.__lock: # prevent multiple threads to write at the same time
            self.target(string)

    def __str__(self) -> str:
        return self.__name

    def __repr__(self) -> str:
        return f"Target({self.__name})"

    def __getitem__(self, key: str) -> Any:
        return self.properties[key]

    def __setitem__(self, key: str, value: Any):
        self.properties[key] = value

    def __delitem__(self, key: str):
        del self.properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.properties

    @property
    def type(self) -> 'Target.Type':
        return self.__type

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name : str):
        old_name = self.__name
        self.__name = name
        del Target.__instances[old_name]
        Target.__instances[name] = self

    def delete(self):
        Target.unregister(self)


    @staticmethod
    def get(name : str | TERMINAL_TARGETS) -> 'Target':
        name = str(name)
        if Target.exist(name):
            return Target.__instances[name]
        else:
            raise ValueError(f"Target {name} does not exist")

    @staticmethod
    def exist(name : str | TERMINAL_TARGETS) -> bool:
        name = str(name)
        return name in Target.__instances.keys()

    @staticmethod
    def list() -> list['Target']:
        return list(Target.__instances.values())

    @staticmethod
    def clear():
        Target.__instances = {}

    @staticmethod
    def register(target : 'Target'):
        Target.__instances[target.name] = target

    @staticmethod
    def unregister(target):
        name = target if isinstance(target, str) else target.name
        if Target.exist(name):
            del Target.__instances[name]
        else:
            raise ValueError(f"Target {name} does not exist")


class LoggerConfig:
    def __init__(self, sensitiveDatas: list[str] = None):
        if sensitiveDatas is None:
            sensitiveDatas = []
        self.sensitiveDatas = sensitiveDatas
        self.showThreadsName = False
        self.showProcessName = False


    def clear(self):
        self.sensitiveDatas = []
        self.showThreadsName = False
        self.showProcessName = False


    def __getitem__(self, key: str) -> Any:
        match key:
            case 'sensitiveDatas':
                return self.sensitiveDatas
            case 'showThreadsName':
                return self.showThreadsName
            case 'showProcessName':
                return self.showProcessName
            case _:
                raise KeyError(f"Parameter {key} not found")

    def __setitem__(self, key: str, value: Any):
        match key:
            case 'sensitiveDatas':
                self.sensitiveDatas = value
            case 'showThreadsName':
                self.showThreadsName = value
            case 'showProcessName':
                self.showProcessName = value
            case _:
                raise KeyError(f"Parameter {key} not found")

    def __str__(self):
        return f"LoggerConfig(sensitiveDatas={self.sensitiveDatas}, showThreadsName={self.showThreadsName}, showProcessName={self.showProcessName})"



from .gamuLogger import (LEVELS, SENSITIVE_LEVELS, TERMINAL_TARGETS, Logger,
                         Module, Target, chrono, critical, debug, debugFunc,
                         deepDebug, deepDebugFunc, error, info, message,
                         warning)
